'''
   MFEM example 17 

   How to run:
       python <arguments>

   Example of arguments:
       ex17.py -m beam-tri.mesh
       ex17.py -m beam-quad.mesh
       ex17.py -m beam-tet.mesh
       ex17.py -m beam-hex.mesh
       ex17.py -m beam-quad.mesh -r 2 -o 3
       ex17.py -m beam-quad.mesh -r 2 -o 2 -a 1 -k 1
       ex17.py -m beam-hex.mesh -r 2 -o 2

'''
import sys
from mfem.common.arg_parser import ArgParser
from os.path import expanduser, join, dirname
import numpy as np
from mfem import path

import mfem.ser as mfem
from mfem.ser import intArray

#


class InitDisplacement(mfem.VectorPyCoefficient):
    def __init__(self, dim):
        self.dim = dim
        mfem.VectorPyCoefficient.__init__(self, dim)

    def EvalValue(self, x):
        u = [0.0]*dim
        u[-1] = -0.2*x[0]
        return tuple(u)


class StressCoefficient(mfem.PyCoefficientBase):
    def __init__(self, lambda_, mu_, si=0, sj=0):
        super(StressCoefficient, self).__init__(0)
        self.lam = lambda_   # coefficient
        self.mu = mu_       # coefficient
        self.si = si
        self.sj = sj     # component
        self.u = None   # displacement GridFunction
        self.grad = mfem.DenseMatrix()

    def SetComponent(self, i, j):
        self.si = i
        self.sj = j

    def SetDisplacement(self, u):
        self.u = u

    def Eval(self, T, ip):
        si, sj = self.si, self.sj
        L = self.lam.Eval(T, ip)
        M = self.mu.Eval(T, ip)
        self.u.GetVectorGradient(T, self.grad)
        if (self.si == self.sj):
            div_u = self.grad.Trace()
            return L * div_u + 2 * M * self.grad[si, si]
        else:
            return M * (self.grad[si, sj] + self.grad[sj, si])


class VisMan(object):
    def __init__(self, vishost, visport):
        self.host = vishost
        self.port = visport
        self.socks = []
        self.output = None
        self.win_x = 0
        self.win_y = 0
        self.win_w = 200  # window width
        self.win_h = 150  # window height
        self.stride_x = self.win_w
        self.stride_y = self.win_h + 20
        self.win_nx = 4  # number of windows in a row
        self.sid = 0

    def NewWindow(self):
        self.socks.append(mfem.socketstream(self.host, self.port))
        self.output = self.socks[-1]
        self.output.precision(8)
        self.socks
        self.sid = self.sid + 1

    def CloseConnection(self):
        self.socks = []
        del self.output
        self.output = None

    def PositionWindow(self):
        if self.output is None:
            return

        sid = self.sid
        command = ("window_geometry " +
                   str(self.win_x + self.stride_x*(sid % self.win_nx)) +
                   ' ' +
                   str(self.win_y + self.stride_y*(sid/self.win_nx)) +
                   ' ' + str(self.win_w) + ' ' + str(self.win_h))
        self.output.send_text(command)
        self.output.flush()

    def send_solution(self, mesh, x):
        if self.output is None:
            return
        self.output.send_solution(mesh, x)

    def send_text(self, x):
        if self.output is None:
            return
        self.output.send_text(x)

    def flush(self):
        if self.output is None:
            return
        self.output.flush()


parser = ArgParser(description='Ex17')
parser.add_argument('-m', '--mesh',
                    default='beam-tri.mesh',
                    action='store', type=str,
                    help='Mesh file to use.')
parser.add_argument('-r', '--refine',
                    action='store', default=-1, type=int,
                    help="Number of times to refine the mesh uniformly, -1 for auto.")
parser.add_argument('-o', '--order',
                    action='store', default=1, type=int,
                    help="Finite element order (polynomial degree)")
parser.add_argument('-a', '--alpha',
                    action='store', default=-1.0, type=float,
                    help='\n'.join(["One of the two DG penalty parameters, typically +1/-1."
                                    " See the documentation of class DGElasticityIntegrator."]))
parser.add_argument('-k', '--kappa',
                    action='store', default=-1.0, type=float,
                    help='\n'.join(["One of the two DG penalty parameters, should be positve."
                                    " Negative values are replaced with (order+1)^2."]))
parser.add_argument('-vis', '--visualization',
                    action='store_true',
                    help='Enable GLVis visualization')

args = parser.parse_args()
ref_levels = args.refine
order = args.order
alpha = args.alpha
kappa = args.kappa
visualization = args.visualization
if (kappa < 0):
    kappa = (order+1.)*(order+1.)
    args.kappa = kappa
parser.print_options(args)
# 2. Read the mesh from the given mesh file
meshfile = expanduser(
    join(dirname(__file__), '..', 'data', args.mesh))
mesh = mfem.Mesh(meshfile, 1, 1)
dim = mesh.Dimension()
if (mesh.attributes.Max() < 2 or
        mesh.bdr_attributes.Max() < 2):
    print("\n".join(["Input mesh should have at least two materials and ",
                     "two boundary attributes! (See schematic in ex17.cpp)\n"]))
    sys.exit()

# 3. Refine the mesh to increase the resolution.
ref_levels = int(np.floor(np.log(5000./mesh.GetNE())/np.log(2.)/dim))
for x in range(ref_levels):
    mesh.UniformRefinement()

# Since NURBS meshes do not support DG integrators, we convert them to
# regular polynomial mesh of the specified (solution) order.
if (mesh.NURBSext):
    mesh.SetCurvature(order)

# 4. Define a DG vector finite element space on the mesh. Here, we use
#    Gauss-Lobatto nodal basis because it gives rise to a sparser matrix
#    compared to the default Gauss-Legendre nodal basis.
fec = mfem.DG_FECollection(order, dim, mfem.BasisType.GaussLobatto)
fespace = mfem.FiniteElementSpace(mesh, fec, dim)
print('Number of finite element unknowns: ' + str(fespace.GetVSize()))
print('Assembling:')

# 5. In this example, the Dirichlet boundary conditions are defined by
#    marking boundary attributes 1 and 2 in the marker Array 'dir_bdr'.
#    These b.c. are imposed weakly, by adding the appropriate boundary
#    integrators over the marked 'dir_bdr' to the bilinear and linear forms.
#    With this DG formulation, there are no essential boundary conditions.
ess_tdof_list = intArray()
dir_bdr = intArray(mesh.bdr_attributes.Max())
dir_bdr.Assign(0)
dir_bdr[0] = 1  # boundary attribute 1 is Dirichlet
dir_bdr[1] = 1  # boundary attribute 2 is Dirichlet

# 6. Define the DG solution vector 'x' as a finite element grid function
#    corresponding to fespace. Initialize 'x' using the 'InitDisplacement'
#    function.
x = mfem.GridFunction(fespace)
init_x = InitDisplacement(dim)
x.ProjectCoefficient(init_x)

# 7. Set up the Lame constants for the two materials. They are defined as
#    piece-wise (with respect to the element attributes) constant
#    coefficients, i.e. type PWConstCoefficient.
lamb = mfem.Vector(mesh.attributes.Max())  # lambda is not possible in python
lamb.Assign(1.0)
lamb[0] = 50.
lambda_c = mfem.PWConstCoefficient(lamb)
mu = mfem.Vector(mesh.attributes.Max())
mu.Assign(1.0)
mu[0] = 50.0
mu_c = mfem.PWConstCoefficient(mu)

# 8. Set up the linear form b(.) which corresponds to the right-hand side of
#    the FEM linear system. In this example, the linear form b(.) consists
#    only of the terms responsible for imposing weakly the Dirichlet
#    boundary conditions, over the attributes marked in 'dir_bdr'. The
#    values for the Dirichlet boundary condition are taken from the
#    VectorFunctionCoefficient 'x_init' which in turn is based on the
#    function 'InitDisplacement'.
b = mfem.LinearForm(fespace)
print('r.h.s ...')
integrator = mfem.DGElasticityDirichletLFIntegrator(
    init_x, lambda_c, mu_c, alpha, kappa)
b.AddBdrFaceIntegrator(integrator, dir_bdr)
b.Assemble()

# 9. Set up the bilinear form a(.,.) on the DG finite element space
#    corresponding to the linear elasticity integrator with coefficients
#    lambda and mu as defined above. The additional interior face integrator
#    ensures the weak continuity of the displacement field. The additional
#    boundary face integrator works together with the boundary integrator
#    added to the linear form b(.) to impose weakly the Dirichlet boundary
#    conditions.
a = mfem.BilinearForm(fespace)
a.AddDomainIntegrator(mfem.ElasticityIntegrator(lambda_c, mu_c))

a.AddInteriorFaceIntegrator(
    mfem.DGElasticityIntegrator(lambda_c, mu_c, alpha, kappa))
a.AddBdrFaceIntegrator(mfem.DGElasticityIntegrator(
    lambda_c, mu_c, alpha, kappa), dir_bdr)
print('matrix ...')
a.Assemble()

A = mfem.SparseMatrix()
B = mfem.Vector()
X = mfem.Vector()
a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)
print('...done')


A.PrintInfo()
'''
   Note: extension of ostream &

   A.PrintInfo()               # output to std::cout
   A.PrintInfo("matrix info")  # output to file 

   Above two are the same as

   from mfem._ser.io_stream import STDOUT, wFILE
   A.PrintInfo(STDOUT)
   A.PrintInfo(wFILE("matrix_info"))

'''

# 11. Define a simple symmetric Gauss-Seidel preconditioner and use it to
#     solve the system Ax=b with PCG for the symmetric formulation, or GMRES
#     for the non-symmetric.
M = mfem.GSSmoother(A)
rtol = 1e-6
if (alpha == -1.0):
    mfem.PCG(A, M, B, X, 3, 5000, rtol*rtol, 0.0)
else:
    mfem.GMRES(A, M, B, X, 3, 5000, 50, rtol*rtol, 0.0)

# 12. Recover the solution as a finite element grid function 'x'.
a.RecoverFEMSolution(X, b, x)

# 13. Use the DG solution space as the mesh nodal space. This allows us to
#     save the displaced mesh as a curved DG mesh.
mesh.SetNodalFESpace(fespace)
reference_nodes = mfem.Vector()
if (visualization):
    reference_nodes.Assign(mesh.GetNodes())
# 14. Save the displaced mesh and minus the solution (which gives the
#     backward displacements to the reference mesh). This output can be
#     viewed later using GLVis: "glvis -m displaced.mesh -g sol.gf".
nodes = mesh.GetNodes()
nodes += x
x.Neg()

mesh.Print('displaced.mesh', 8)
x.Save('sol.gf', 8)

# 15. Visualization: send data by socket to a GLVis server.
if (visualization):
    vis = VisMan("localhost", 19916)
    glvis_keys = "Rjlc" if (dim < 3) else "c"

    vis.NewWindow()
    vis.send_solution(mesh, x)
    vis.send_text("keys " + glvis_keys)
    vis.send_text("window_title 'Deformed configuration'")
    vis.send_text("plot_caption 'Backward displacement'")
    vis.PositionWindow()
    vis.CloseConnection()

    c = "xyz"
    scalar_dg_space = mfem.FiniteElementSpace(mesh, fec)
    stress = mfem.GridFunction(scalar_dg_space)
    stress_c = StressCoefficient(lambda_c, mu_c)

    mesh.GetNodes().Assign(reference_nodes)
    x.Neg()
    stress_c.SetDisplacement(x)

    def make_plot(si, sj):
        stress_c.SetComponent(si, sj)
        stress.ProjectCoefficient(stress_c)
        vis.NewWindow()
        vis.send_solution(mesh, stress)
        vis.send_text("keys " + glvis_keys)
        vis.send_text("window_title |Stress" + c[si] + c[sj] + "|")
        vis.PositionWindow()
        vis.CloseConnection()

    for si in range(dim):
        for jj in range(dim-si):
            make_plot(si, si+jj)



'''
   MFEM example 23
      See c++ version in the MFEM library for more detail 
'''
import os
import io
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp, sqrt, pi, abs, array, floor, log, sum


def run(mesh_file="",
        ref_levels=2,
        order=2,
        ode_solver_type=10,
        t_final=0.5,
        dt=1e-2,
        speed=1.0,
        dirichlet=True,
        visit=True,
        visualization=True,
        vis_steps=5):

    class WaveOperator(mfem.SecondOrderTimeDependentOperator):
        def __init__(self, fespace, ess_bdr, speed):
            mfem.SecondOrderTimeDependentOperator.__init__(
                self, fespace.GetTrueVSize(), 0.0)

            self.ess_tdof_list = mfem.intArray()

            rel_tol = 1e-8
            fespace.GetEssentialTrueDofs(ess_bdr, self.ess_tdof_list)

            c2 = mfem.ConstantCoefficient(speed*speed)
            K = mfem.BilinearForm(fespace)
            K.AddDomainIntegrator(mfem.DiffusionIntegrator(c2))
            K.Assemble()

            self.Kmat0 = mfem.SparseMatrix()
            self.Kmat = mfem.SparseMatrix()
            dummy = mfem.intArray()
            K.FormSystemMatrix(dummy, self.Kmat0)
            K.FormSystemMatrix(self.ess_tdof_list, self.Kmat)
            self.K = K

            self.Mmat = mfem.SparseMatrix()
            M = mfem.BilinearForm(fespace)
            M.AddDomainIntegrator(mfem.MassIntegrator())
            M.Assemble()
            M.FormSystemMatrix(self.ess_tdof_list, self.Mmat)
            self.M = M

            M_solver = mfem.CGSolver()
            M_prec = mfem.DSmoother()
            M_solver.iterative_mode = False
            M_solver.SetRelTol(rel_tol)
            M_solver.SetAbsTol(0.0)
            M_solver.SetMaxIter(30)
            M_solver.SetPrintLevel(0)
            M_solver.SetPreconditioner(M_prec)
            M_solver.SetOperator(self.Mmat)
            self.M_prec = M_prec
            self.M_solver = M_solver

            T_solver = mfem.CGSolver()
            T_prec = mfem.DSmoother()
            T_solver.iterative_mode = False
            T_solver.SetRelTol(rel_tol)
            T_solver.SetAbsTol(0.0)
            T_solver.SetMaxIter(100)
            T_solver.SetPrintLevel(0)
            T_solver.SetPreconditioner(T_prec)
            self.T_prec = T_prec
            self.T_solver = T_solver
            self.T = None

        def Mult(self, u, du_dt, d2udt2):
            # Compute:
            #    d2udt2 = M^{-1}*-K(u)
            # for d2udt2
            z = mfem.Vector(u.Size())
            self.Kmat.Mult(u, z)
            z.Neg()  # z = -z
            self.M_solver.Mult(z, d2udt2)

        def ImplicitSolve(self, fac0, fac1, u, dudt, d2udt2):
            # Solve the equation:
            #    d2udt2 = M^{-1}*[-K(u + fac0*d2udt2)]
            # for d2udt2
            if self.T is None:
                self.T = mfem.Add(1.0, self.Mmat, fac0, self.Kmat)
                self.T_solver.SetOperator(self.T)
            z = mfem.Vector(u.Size())
            self.Kmat0.Mult(u, z)
            z.Neg()

            # iterate over Array<int> :D
            for j in self.ess_tdof_list:
                z[j] = 0.0

            self.T_solver.Mult(z, d2udt2)

        def SetParameters(self, u):
            self.T = None

    mesh = mfem.Mesh(mesh_file, 1, 1)
    dim = mesh.Dimension()

    # 3. Define the ODE solver used for time integration. Several second order
    #    time integrators are available.
    if ode_solver_type <= 10:
        ode_solver = mfem.GeneralizedAlpha2Solver(ode_solver_type/10)
    elif ode_solver_type == 11:
        ode_solver = mfem.AverageAccelerationSolver()
    elif ode_solver_type == 12:
        ode_solver = mfem.LinearAccelerationSolver()
    elif ode_solver_type == 13:
        ode_solver = mfem.CentralDifferenceSolver()
    elif ode_solver_type == 14:
        ode_solver = mfem.FoxGoodwinSolver()
    else:
        print("Unknown ODE solver type: " + str(ode_solver_type))

    # 4. Refine the mesh to increase the resolution. In this example we do
    #    'ref_levels' of uniform refinement, where 'ref_levels' is a
    #    command-line parameter.
    for lev in range(ref_levels):
        mesh.UniformRefinement()

    # 5. Define the vector finite element space representing the current and the
    #    initial temperature, u_ref.
    fe_coll = mfem.H1_FECollection(order, dim)
    fespace = mfem.FiniteElementSpace(mesh, fe_coll)

    fe_size = fespace.GetTrueVSize()
    print("Number of temperature unknowns: " + str(fe_size))

    u_gf = mfem.GridFunction(fespace)
    dudt_gf = mfem.GridFunction(fespace)

    # 6. Set the initial conditions for u. All boundaries are considered
    #    natural.
    class cInitialSolution(mfem.PyCoefficient):
        def EvalValue(self, x):
            norm2 = sum(x**2)
            return exp(-norm2*30)

    class cInitialRate(mfem.PyCoefficient):
        def EvalValue(self, x):
            return 0

    u_0 = cInitialSolution()
    dudt_0 = cInitialRate()

    u_gf.ProjectCoefficient(u_0)
    u = mfem.Vector()
    u_gf.GetTrueDofs(u)

    dudt_gf.ProjectCoefficient(dudt_0)
    dudt = mfem.Vector()
    dudt_gf.GetTrueDofs(dudt)

    # 7. Initialize the conduction operator and the visualization.
    ess_bdr = mfem.intArray()
    if mesh.bdr_attributes.Size():
        ess_bdr.SetSize(mesh.bdr_attributes.Max())
        if (dirichlet):
            ess_bdr.Assign(1)
        else:
            ess_bdr.Assigne(0)

    oper = WaveOperator(fespace, ess_bdr, speed)

    u_gf.SetFromTrueDofs(u)

    mesh.Print("ex23.mesh", 8)
    output = io.StringIO()
    output.precision = 8
    u_gf.Save(output)
    dudt_gf.Save(output)
    fid = open("ex23-init.gf", 'w')
    fid.write(output.getvalue())
    fid.close()

    if visit:
        visit_dc = mfem.VisItDataCollection("Example23", mesh)
        visit_dc.RegisterField("solution", u_gf)
        visit_dc.RegisterField("rate", dudt_gf)
        visit_dc.SetCycle(0)
        visit_dc.SetTime(0.0)
        visit_dc.Save()

    if visualization:
        sout = mfem.socketstream("localhost", 19916)
        if not sout.good():
            print("Unable to connect to GLVis server at localhost:19916")
            visualization = False
            print("GLVis visualization disabled.")
        else:
            sout.precision(precision)
            sout << "solution\n" << mesh << dudt_gf
            sout << "pause\n"
            sout.flush()
            print(
                "GLVis visualization paused. Press space (in the GLVis window) to resume it.")

    # 8. Perform time-integration (looping over the time iterations, ti, with a
    #    time-step dt).
    ode_solver.Init(oper)
    t = 0.0

    last_step = False
    ti = 0
    while not last_step:
        ti += 1
        if t + dt >= t_final - dt/2:
            last_step = True

        t, dt = ode_solver.Step(u, dudt, t, dt)

        if last_step or (ti % vis_steps == 0):
            print("step " + str(ti) + ", t = " + "{:g}".format(t))

            u_gf.SetFromTrueDofs(u)
            dudt_gf.SetFromTrueDofs(dudt)
            if visualization:
                sout << "solution\n" << mesh << u_gf
                sout.flush()

            if visit:
                visit_dc.SetCycle(ti)
                visit_dc.SetTime(t)
                visit_dc.Save()

        oper.SetParameters(u)

    # 9. Save the final solution. This output can be viewed later using GLVis:
    #    "glvis -m ex23.mesh -g ex23-final.gf".
    output = io.StringIO()
    output.precision = 8
    u_gf.Save(output)
    dudt_gf.Save(output)
    fid = open("ex23-final.gf", "w")
    fid.write(output.getvalue())
    fid.close()


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser
    parser = ArgParser(description="Ex23 (Wave problem)")
    parser.add_argument('-m', '--mesh',
                        default='star.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-r', '--refine',
                        action='store', default=2, type=int,
                        help="Number of times to refine the mesh uniformly before parallel")
    parser.add_argument('-o', '--order',
                        action='store', default=2, type=int,
                        help="Finite element order (polynomial degree)")
    help_ode = '\n'.join(["ODE solver: [0--10] \t- GeneralizedAlpha(0.1 * s),",
                          "11 \t - Average Acceleration,",
                          "12 \t - Linear Acceleration",
                          "13 \t- CentralDifference",
                          "14 \t- FoxGoodwin"])
    parser.add_argument('-s', '--ode-solver',
                        action='store', default=10, type=int,
                        help=help_ode)
    parser.add_argument('-tf', '--t-final',
                        action='store', default=0.5, type=float,
                        help="Final time; start time is 0.")
    parser.add_argument('-dt', '--time-step',
                        action='store', default=1e-2, type=float,
                        help="Time step")
    parser.add_argument("-c", "--speed",
                        action='store', default=1.0, type=float,
                        help="Wave speed.")
    parser.add_argument("-neu",  "--neumann",
                        action='store_true', default=False,
                        help="BC switch.")
    parser.add_argument('-vis', '--visualization',
                        action='store_true', default=True,
                        help='Enable GLVis visualization')
    parser.add_argument('-visit', '--visit-datafiles',
                        action='store_true', default=True,
                        help="Save data files for VisIt (visit.llnl.gov) visualization.")
    parser.add_argument("-vs", "--visualization-steps",
                        action='store', default=5,  type=int,
                        help="Visualize every n-th timestep.")

    args = parser.parse_args()
    parser.print_options(args)
    mesh_file = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))

    run(mesh_file=mesh_file,
        ref_levels=args.refine,
        order=args.order,
        ode_solver_type=args.ode_solver,
        t_final=args.t_final,
        dt=args.time_step,
        speed=args.speed,
        dirichlet=(not args.neumann),
        visit=args.visit_datafiles,
        vis_steps=args.visualization_steps,
        visualization=args.visualization)



import mfem

if mfem.mfem_mode is None:
   mfem.mfem_mode = 'parallel'
if mfem.mfem_mode == 'serial':
   raise ImportError("MFEM serial mode is already loaded")
debug_print = mfem.debug_print

from mpi4py import MPI




'''
   MFEM example 19

      This examples solves a quasi-static incompressible nonlinear
      elasticity problem of the form 0 = H(x), where H is an
      incompressible hyperelastic model and x is a block state vector
      containing displacement and pressure variables. 

      See c++ version in the MFEM library for more detail 

   Sample runs:
      python ex19.py -m ../data/beam-quad.mesh
      python ex19.py -m ../data/beam-tri.mesh
      python ex19.py -m ../data/beam-hex.mesh
      python ex19.py -m ../data/beam-tet.mesh

'''


import sys
from mfem.common.arg_parser import ArgParser
import mfem.ser as mfem
from mfem.ser import intArray, add_vector, Add
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sqrt, pi, cos, sin, hypot, arctan2
from scipy.special import erfc


parser = ArgParser(description='Ex19')
parser.add_argument('-m', '--mesh',
                    default='beam-tet.mesh',
                    action='store', type=str,
                    help='Mesh file to use.')
parser.add_argument('-r', '--refine',
                    action='store', default=0, type=int,
                    help="Number of times to refine the mesh uniformly")
parser.add_argument('-o', '--order',
                    action='store', default=2, type=int,
                    help="Finite element order (polynomial degree)")
parser.add_argument('-rel', '--relative-tolerance',
                    action='store', default=1e-4, type=float,
                    help="Relative tolerance for the Newton solve.")
parser.add_argument('-abs', '--absolute-tolerance',
                    action='store', default=1e-6, type=float,
                    help="Absolute tolerance for the Newton solve.")
parser.add_argument("-it", "--newton-iterations",
                    action='store', default=500, type=int,
                    help="Maximum iterations for the Newton solve.")
parser.add_argument("-mu", "--shear-modulus",
                    action='store', default=1.0, type=float,
                    help="Shear modulus in the Neo-Hookean material.")
parser.add_argument('-vis', '--visualization',
                    action='store_true', default=True,
                    help='Enable GLVis visualization')

args = parser.parse_args()


def ex19_main(args):
    ref_levels = args.refine
    order = args.order
    visualization = args.visualization
    mu = args.shear_modulus
    newton_rel_tol = args.relative_tolerance
    newton_abs_tol = args.absolute_tolerance
    newton_iter = args.newton_iterations

    parser.print_options(args)

    meshfile = expanduser(join(dirname(__file__), '..', 'data', args.mesh))
    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()

    for lev in range(ref_levels):
        mesh.UniformRefinement()

    #  4. Define the shear modulus for the incompressible Neo-Hookean material
    c_mu = mfem.ConstantCoefficient(mu)

    #  5. Define the finite element spaces for displacement and pressure
    #     (Taylor-Hood elements). By default, the displacement (u/x) is a second
    #     order vector field, while the pressure (p) is a linear scalar function.
    quad_coll = mfem.H1_FECollection(order, dim)
    lin_coll = mfem.H1_FECollection(order-1, dim)

    R_space = mfem.FiniteElementSpace(
        mesh, quad_coll, dim, mfem.Ordering.byVDIM)
    W_space = mfem.FiniteElementSpace(mesh, lin_coll)

    spaces = [R_space, W_space]
    R_size = R_space.GetVSize()
    W_size = W_space.GetVSize()

    #   6. Define the Dirichlet conditions (set to boundary attribute 1 and 2)
    ess_bdr_u = mfem.intArray(R_space.GetMesh().bdr_attributes.Max())
    ess_bdr_p = mfem.intArray(W_space.GetMesh().bdr_attributes.Max())
    ess_bdr_u.Assign(0)
    ess_bdr_u[0] = 1
    ess_bdr_u[1] = 1
    ess_bdr_p.Assign(0)
    ess_bdr = [ess_bdr_u, ess_bdr_p]

    print("***********************************************************")
    print("dim(u) = " + str(R_size))
    print("dim(p) = " + str(W_size))
    print("dim(u+p) = " + str(R_size + W_size))
    print("***********************************************************")

    block_offsets = intArray([0, R_size, W_size])
    block_offsets.PartialSum()

    xp = mfem.BlockVector(block_offsets)

    #  9. Define grid functions for the current configuration, reference
    #     configuration, final deformation, and pressure
    x_gf = mfem.GridFunction(R_space)
    x_ref = mfem.GridFunction(R_space)
    x_def = mfem.GridFunction(R_space)
    p_gf = mfem.GridFunction(W_space)

    x_gf.MakeRef(R_space, xp.GetBlock(0), 0)
    p_gf.MakeRef(W_space, xp.GetBlock(1), 0)

    deform = InitialDeformation(dim)
    refconfig = ReferenceConfiguration(dim)

    x_gf.ProjectCoefficient(deform)
    x_ref.ProjectCoefficient(refconfig)
    p_gf.Assign(0.0)

    #  10. Initialize the incompressible neo-Hookean operator
    oper = RubberOperator(spaces, ess_bdr, block_offsets,
                          newton_rel_tol, newton_abs_tol, newton_iter,
                          mu)
    #  11. Solve the Newton system
    oper.Solve(xp)

    #  12. Compute the final deformation
    mfem.subtract_vector(x_gf, x_ref, x_def)

    #  13. Visualize the results if requested
    if (visualization):
        vis_u = mfem.socketstream("localhost", 19916)
        visualize(vis_u, mesh, x_gf, x_def, "Deformation", True)
        vis_p = mfem.socketstream("localhost", 19916)
        visualize(vis_p, mesh, x_gf, p_gf, "Deformation", True)

    #  14. Save the displaced mesh, the final deformation, and the pressure
    nodes = x_gf
    owns_nodes = 0
    nodes, owns_nodes = mesh.SwapNodes(nodes, owns_nodes)

    mesh.Print('deformed.mesh', 8)
    p_gf.Save('pressure.sol', 8)
    x_def.Save("deformation.sol",  8)


'''
 Custom block preconditioner for the Jacobian of the incompressible nonlinear
 elasticity operator. It has the form

 P^-1 = [ K^-1 0 ][ I -B^T ][ I  0           ]
        [ 0    I ][ 0  I   ][ 0 -\gamma S^-1 ]

 where the original Jacobian has the form

 J = [ K B^T ]
     [ B 0   ]

 and K^-1 is an approximation of the inverse of the displacement part of the
 Jacobian and S^-1 is an approximation of the inverse of the Schur
 complement S = B K^-1 B^T. The Schur complement is approximated using
 a mass matrix of the pressure variables.
'''


class JacobianPreconditioner(mfem.Solver):
    def __init__(self, spaces, mass, offsets):
        self.pressure_mass = mass
        self.block_offsets = offsets
        super(JacobianPreconditioner, self).__init__(offsets[2])

        self.gamma = 0.00001

        # The mass matrix and preconditioner do not change every Newton cycle, so we
        # only need to define them once
        self.mass_prec = mfem.GSSmoother(mass)

        mass_pcg = mfem.CGSolver()
        mass_pcg.SetRelTol(1e-12)
        mass_pcg.SetAbsTol(1e-12)
        mass_pcg.SetMaxIter(200)
        mass_pcg.SetPrintLevel(0)
        mass_pcg.SetPreconditioner(self.mass_prec)
        mass_pcg.SetOperator(self.pressure_mass)
        mass_pcg.iterative_mode = False
        self.mass_pcg = mass_pcg

        # The stiffness matrix does change every Newton cycle, so we will define it
        # during SetOperator
        self.stiff_pcg = None
        self.stiff_prec = None

    def Mult(self, k, y):
        # Extract the blocks from the input and output vectors
        block_offsets = self.block_offsets
        disp_in = k[block_offsets[0]:block_offsets[1]]
        pres_in = k[block_offsets[1]:block_offsets[2]]

        disp_out = y[block_offsets[0]:block_offsets[1]]
        pres_out = y[block_offsets[1]:block_offsets[2]]

        temp = mfem.Vector(block_offsets[1]-block_offsets[0])
        temp2 = mfem.Vector(block_offsets[1]-block_offsets[0])

        # Perform the block elimination for the preconditioner
        self.mass_pcg.Mult(pres_in, pres_out)
        pres_out *= -self.gamma

        self.jacobian.GetBlock(0, 1).Mult(pres_out, temp)
        mfem.subtract_vector(disp_in, temp, temp2)
        self.stiff_pcg.Mult(temp2, disp_out)

    def SetOperator(self, op):
        self.jacobian = mfem.Opr2BlockOpr(op)
        if (self.stiff_prec == None):
            # Initialize the stiffness preconditioner and solver
            stiff_prec_gs = mfem.GSSmoother()
            self.stiff_prec = stiff_prec_gs

            stiff_pcg_iter = mfem.GMRESSolver()
            stiff_pcg_iter.SetRelTol(1e-8)
            stiff_pcg_iter.SetAbsTol(1e-8)
            stiff_pcg_iter.SetMaxIter(200)
            stiff_pcg_iter.SetPrintLevel(0)
            stiff_pcg_iter.SetPreconditioner(self.stiff_prec)
            stiff_pcg_iter.iterative_mode = False

            self.stiff_pcg = stiff_pcg_iter

        #  At each Newton cycle, compute the new stiffness preconditioner by updating
        #  the iterative solver which, in turn, updates its preconditioner
        self.stiff_pcg.SetOperator(self.jacobian.GetBlock(0, 0))


class GeneralResidualMonitor(mfem.IterativeSolverMonitor):
    def __init__(self, prefix, print_level):
        mfem.IterativeSolverMonitor.__init__(self)
        self.print_level = print_level
        self.prefix = prefix
        self.norm = 0

    def MonitorResidual(self, it, norm, r, final):
        txt = ''
        if (self.print_level == 1 or
                (self.print_level == 3 and (final or it == 0))):
            txt = (txt + self.prefix + " iteration " + '{:>2}'.format(it) +
                   " : ||r|| = " + "{:g}".format(norm))
            if it > 0:
                txt = txt + ",  ||r||/||r_0|| = " + \
                    "{:g}".format(norm/self.norm0)
            else:
                self.norm0 = norm
            print(txt)


class RubberOperator(mfem.PyOperator):
    def __init__(self, spaces, ess_bdr, block_offsets,
                 rel_tol, abs_tol, iter, mu):

        # Array<Vector *> -> tuple
        super(RubberOperator, self).__init__(
            spaces[0].GetVSize() + spaces[1].GetVSize())
        rhs = (None, None)

        self.spaces = spaces
        self.mu = mfem.ConstantCoefficient(mu)
        self.block_offsets = block_offsets

        Hform = mfem.BlockNonlinearForm(spaces)
        Hform.AddDomainIntegrator(
            mfem.IncompressibleNeoHookeanIntegrator(self.mu))
        Hform.SetEssentialBC(ess_bdr, rhs)
        self.Hform = Hform

        a = mfem.BilinearForm(self.spaces[1])
        one = mfem.ConstantCoefficient(1.0)
        a.AddDomainIntegrator(mfem.MassIntegrator(one))
        a.Assemble()
        a.Finalize()
        pressure_mass = a.LoseMat()

        self.j_prec = JacobianPreconditioner(
            spaces, pressure_mass, block_offsets)

        j_gmres = mfem.GMRESSolver()
        j_gmres.iterative_mode = False
        j_gmres.SetRelTol(1e-12)
        j_gmres.SetAbsTol(1e-12)
        j_gmres.SetMaxIter(300)
        j_gmres.SetPrintLevel(-1)
        self.j_monitor = GeneralResidualMonitor("  GMRES", 3)
        j_gmres.SetMonitor(self.j_monitor)
        j_gmres.SetPreconditioner(self.j_prec)
        self.j_solver = j_gmres

        newton_solver = mfem.NewtonSolver()
        # Set the newton solve parameters
        newton_solver.iterative_mode = True
        newton_solver.SetSolver(self.j_solver)
        newton_solver.SetOperator(self)
        newton_solver.SetPrintLevel(-1)
        newton_solver.SetRelTol(rel_tol)
        newton_solver.SetAbsTol(abs_tol)
        newton_solver.SetMaxIter(iter)
        self.n_monitor = GeneralResidualMonitor("Newton", 1)
        newton_solver.SetMonitor(self.n_monitor)
        self.newton_solver = newton_solver

    def Solve(self, xp):
        zero = mfem.Vector()
        self.newton_solver.Mult(zero, xp)
        if not self.newton_solver.GetConverged():
            assert False, "Newton Solver did not converge."

    def Mult(self, k, y):
        self.Hform.Mult(k, y)

    def GetGradient(self, xp):
        return self.Hform.GetGradient(xp)


#  Inline visualization
def visualize(out, mesh, deformed_nodes, field, field_name, init_vis):
    if out is None:
        return

    nodes = deformed_nodes
    owns_nodes = 0

    nodes, owns_nodes = mesh.SwapNodes(nodes, owns_nodes)

    out.precision(8)
    out.send_solution(mesh, field)

    nodes, owns_nodes = mesh.SwapNodes(nodes, owns_nodes)

    if (init_vis):
        out.send_text("window_size 800 800")
        out.send_text("window_title '" + field_name)
        if (mesh.SpaceDimension() == 2):
            out.send_text("view 0 0")
            out.send_text("keys jlA")
        out.send_text("keys cm")         # show colorbar and mesh
        # update value-range; keep mesh-extents fixed
        out.send_text("autoscale value")
    out.flush()

#  FunctionCOefficient is wrapperd usng VectorPyCoefficient class


class InitialDeformation(mfem.VectorPyCoefficient):
    def EvalValue(self, x):
        y = x.copy()
        y[1] = x[1] + 0.25*x[0]
        return y
#  FunctionCOefficient is wrapperd usng VectorPyCoefficient class


class ReferenceConfiguration(mfem.VectorPyCoefficient):
    def EvalValue(self, x):
        return x


if __name__ == '__main__':
    ex19_main(args)



'''
   MFEM example 30

   See c++ version in the MFEM library for more detail 
'''
import mfem.ser as mfem
from mfem.ser import intArray
import os
from os.path import expanduser, join
import numpy as np
from numpy import sin, array

try:
    from numba import jit
except:
    assert False, "This example requires numba"


@mfem.jit.scalar()
def affine_coeff(p):
    if p[0] < 0.0:
        return 1 + p[0] + p[1]
    else:
        return 1


@mfem.jit.scalar()
def jump_coeff(p):
    norm = np.sqrt(p[0]**2 + p[1]**2)
    if norm > 0.4 and norm < 0.6:
        return 1.0
    else:
        return 5.0


@mfem.jit.scalar()
def singular_coeff(p):
    x = p[0]
    y = p[1]
    alpha = 1000.0
    xc = 0.75
    yc = 0.5
    r0 = 0.7
    r = np.sqrt((x - xc)**2.0 + (y - yc)**2)
    num = - (alpha - alpha**3 * (r**2 - r0**2))
    denom = (r * (alpha**2 * r0**2 + alpha**2 * r**2 -
                  2 * alpha**2 * r0 * r + 1.0))**2
    denom = max([denom, 1e-8])
    return num / denom


def run(nc_limit=1,
        order=1,
        meshfile='',
        visualization=True,
        max_elems=100000,
        enriched_order=5,
        osc_threshold=1e-3):

    mesh = mfem.Mesh(meshfile, 1, 1)

    if mesh.NURBSext:
        for i in range(2):
            mesh.UniformRefinement()
        mesh.SetCurvature(2)

    coeffrefiner = mfem.CoefficientRefiner(affine_coeff, order)

    if (visualization):
        sol_sock = mfem.socketstream("localhost", 19916)

    irs = [mfem.IntRules.Get(i, 2*order + enriched_order)
           for i in range(mfem.Geometry.NumGeom)]

    coeffrefiner.SetIntRule(irs)
    coeffrefiner.SetMaxElements(max_elems)
    coeffrefiner.SetThreshold(osc_threshold)
    coeffrefiner.SetNCLimit(nc_limit)
    coeffrefiner.PrintWarnings()

    coeffrefiner.PreprocessMesh(mesh)

    print()
    print("Function 0 (affine)")
    print("Number of Elements " + str(mesh.GetNE()))
    print("Osc error " + "{:g}".format(coeffrefiner.GetOsc()))

    # 8. Preprocess mesh to control osc (jump function).
    coeffrefiner.ResetCoefficient(jump_coeff)
    coeffrefiner.PreprocessMesh(mesh)

    print("")
    print("Function 1 (discontinuous)")
    print("Number of Elements " + str(mesh.GetNE()))
    print("Osc error " + "{:g}".format(coeffrefiner.GetOsc()))

    # 9. Preprocess mesh to control osc (singular function).
    coeffrefiner.ResetCoefficient(singular_coeff)
    coeffrefiner.PreprocessMesh(mesh)

    print("")
    print("Function 2 (singular)")
    print("Number of Elements " + str(mesh.GetNE()))
    print("Osc error " + "{:g}".format(coeffrefiner.GetOsc()))

    if visualization:
        sol_sock.precision(8)
        sol_sock << "mesh\n" << mesh
        sol_sock.flush()


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex30 (Coefficient Refiner)')
    parser.add_argument('-m', '--mesh',
                        default="star.mesh",
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument("-l", "--nc_limit",
                        action='store',
                        type=int,
                        default=1,
                        help="Maximum level of hanging nodes.")
    parser.add_argument('-me', '--max-elems',
                        action='store',
                        type=int, default=100000,
                        help="Stop after reaching this many elements",)
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')
    parser.add_argument("-e", "--error",
                        action='store',
                        default=1e-3,
                        type=float,
                        help="relative data oscillation threshold.")
    parser.add_argument("-eo", "--enriched-order",
                        type=int,
                        default=5,
                        help="Enriched quadrature order.")

    args = parser.parse_args()
    parser.print_options(args)

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))
    print(meshfile)

    run(nc_limit=args.nc_limit,
        order=args.order,
        meshfile=meshfile,
        visualization=args.visualization,
        max_elems=args.max_elems,
        enriched_order=args.enriched_order,
        osc_threshold=args.error)



file = "../_ser/bilininteg.py"

out = ["namespace mfem {"]
fid = open(file, 'r')
for line in fid.readlines():
    if line.startswith("class"):
        cname = (line.split(' ')[1]).split('(')[0]
         
    if line.startswith("    def __init__"):
        pp = ""
        if line.find("*args") != -1:
            pp = "    self._coeff = args"
        elif line.find("own_bfi_") != -1:
            pp = "    if own_bfi_ == 1:  bfi_.thisown = 0"
        elif line.find("integ, own_integ=1") != -1:
            pp = "    if own_integ == 1:  integ.thisown = 0"
        elif line.find("own_integs=1") != -1:
            pp = "    self.own_integs = own_integs"
        elif line.find(", vq)") != -1:
            pp = "    self._coeff = vq"
        elif line.find(", q)") != -1:
            pp = "    self._coeff = q"
        elif line.find(", q, a=1.0)") != -1:
            pp = "    self._coeff = q"
        elif line.find(", q, i") != -1:
            pp = "    self._coeff = q"         
        elif line.find(", sc)") != -1:
            pp = "    self._coeff = sc"
        elif line.find(", vc)") != -1:
            pp = "    self._coeff = vc"
        elif line.find("(self)") != -1: 
            pass
        elif line.find("(self, ir=None)") != -1: 
            pass
        elif line.find("(self, fes, e=1.0)") != -1:
            pass
        else:
            print(cname)
            print(line)
            assert False, "No recipt for this pattern "
        if pp != "":
            out.append("%pythonappend " + cname + "::" + cname + " %{")
            out.append(pp)
            out.append("%}")
            out.append("")            
fid.close()

out.append("%pythonappend SumIntegrator::AddIntegrator %{")
out.append("   if self.own_integs == 1: integ.thisown = 0")
out.append("%}")
out.append("}")            
out.append("")            

fid = open("bilininteg_ext.i", "w")
fid.write("\n".join(out))
fid.close()



'''
   MFEM example 14p 

   How to run:
      mpirun -np 4 python <arguments>

   Example of arguments:
      ex14p.py -m inline-quad.mesh -o 0
      ex14p.py -m star.mesh -o 2
      ex14p.py -m escher.mesh -s 1
      ex14p.py -m fichera.mesh -s 1 -k 1
      ex14p.py -m square-disc-p3.mesh -o 3
      ex14p.py -m square-disc-nurbs.mesh -o 1
      ex14p.py -m disc-nurbs.mesh -rs 4 -o 2 -s 1 -k 0
      ex14p.py -m pipe-nurbs.mesh -o 1
      ex14p.py -m inline-segment.mesh -rs 5
      ex14p.py -m amr-quad.mesh -rs 3
      ex14p.py -m amr-hex.mesh
'''
import sys
from mfem.common.arg_parser import ArgParser
from os.path import expanduser, join, dirname
import numpy as np

import mfem.par as mfem
from mpi4py import MPI

num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank

parser = ArgParser(description='Ex14p')
parser.add_argument('-m', '--mesh',
                    default='star.mesh',
                    action='store', type=str,
                    help='Mesh file to use.')
parser.add_argument('-rs', '--refine-serial',
                    action='store', default=-1, type=int,
                    help="Number of times to refine the mesh uniformly before parallel")
parser.add_argument('-rp', '--refine-parallel',
                    action='store', default=2, type=int,
                    help="Number of times to refine the mesh uniformly after parallel")
parser.add_argument('-o', '--order',
                    action='store', default=1, type=int,
                    help="Finite element order (polynomial degree)")
parser.add_argument('-s', '--sigma',
                    action='store', default=-1.0, type=float,
                    help='\n'.join(["One of the two DG penalty parameters, typically +1/-1."
                                    " See the documentation of class DGDiffusionIntegrator."]))
parser.add_argument('-k', '--kappa',
                    action='store', default=-1.0, type=float,
                    help='\n'.join(["One of the two DG penalty parameters, should be positve."
                                    " Negative values are replaced with (order+1)^2."]))
parser.add_argument('-vis', '--visualization',
                    action='store_true',
                    help='Enable GLVis visualization')
args = parser.parse_args()
ser_ref_levels = args.refine_serial
par_ref_levels = args.refine_parallel
order = args.order
sigma = args.sigma
kappa = args.kappa
visualization = args.visualization
if (kappa < 0):
    kappa = (order+1.)*(order+1.)
    args.kappa = kappa
if (myid == 0):
    parser.print_options(args)

device = mfem.Device('cpu')
if myid == 0:
    device.Print()

# 3. Read the (serial) mesh from the given mesh file on all processors. We
#    can handle triangular, quadrilateral, tetrahedral and hexahedral meshes
#    with the same code. NURBS meshes are projected to second order meshes.
meshfile = expanduser(join(dirname(__file__), '..', 'data', args.mesh))
mesh = mfem.Mesh(meshfile, 1, 1)
dim = mesh.Dimension()

# 4. Refine the serial mesh on all processors to increase the resolution. In
#    this example we do 'ser_ref_levels' of uniform refinement. By default,
#    or if ser_ref_levels < 0, we choose it to be the largest number that
#    gives a final mesh with no more than 10,000 elements.

if ser_ref_levels < 0:
    ser_ref_levels = int(np.floor(np.log(10000./mesh.GetNE())/np.log(2.)/dim))
for x in range(ser_ref_levels):
    mesh.UniformRefinement()

if (mesh.NURBSext):
    mesh.SetCurvature(max(order, 1))

# 5. Define a parallel mesh by a partitioning of the serial mesh. Refine
#    this mesh further in parallel to increase the resolution. Once the
#    parallel mesh is defined, the serial mesh can be deleted.
pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
del mesh
for x in range(par_ref_levels):
    pmesh.UniformRefinement()

# 6. Define a parallel finite element space on the parallel mesh. Here we
#    use discontinuous finite elements of the specified order >= 0.
fec = mfem.DG_FECollection(order, dim)
fespace = mfem.ParFiniteElementSpace(pmesh, fec)
glob_size = fespace.GlobalTrueVSize()
if (myid == 0):
    print('Number of unknowns: ' + str(glob_size))

# 7. Set up the parallel linear form b(.) which corresponds to the
#    right-hand side of the FEM linear system.
b = mfem.ParLinearForm(fespace)
one = mfem.ConstantCoefficient(1.0)
zero = mfem.ConstantCoefficient(0.0)
b.AddDomainIntegrator(mfem.DomainLFIntegrator(one))
b.AddBdrFaceIntegrator(
    mfem.DGDirichletLFIntegrator(zero, one, sigma, kappa))
b.Assemble()

# 8. Define the solution vector x as a parallel finite element grid function
#    corresponding to fespace. Initialize x with initial guess of zero.
x = mfem.ParGridFunction(fespace)
x.Assign(0.0)

# 9. Set up the bilinear form a(.,.) on the finite element space
#    corresponding to the Laplacian operator -Delta, by adding the Diffusion
#    domain integrator and the interior and boundary DG face integrators.
#    Note that boundary conditions are imposed weakly in the form, so there
#    is no need for dof elimination. After serial and parallel assembly we
#    extract the corresponding parallel matrix A.
a = mfem.ParBilinearForm(fespace)
a.AddDomainIntegrator(mfem.DiffusionIntegrator(one))
a.AddInteriorFaceIntegrator(mfem.DGDiffusionIntegrator(one, sigma, kappa))
a.AddBdrFaceIntegrator(mfem.DGDiffusionIntegrator(one, sigma, kappa))
a.Assemble()
a.Finalize()

# 10. Define the parallel (hypre) matrix and vectors representing a(.,.),
#     b(.) and the finite element approximation.
A = a.ParallelAssemble()
B = b.ParallelAssemble()
X = x.ParallelProject()  # HypreParVector

del a
del b

# 11. Depending on the symmetry of A, define and apply a parallel PCG or
#     GMRES solver for AX=B using the BoomerAMG preconditioner from hypre.
amg = mfem.HypreBoomerAMG(A)
if sigma == -1.0:
    pcg = mfem.HyprePCG(A)
    pcg.SetTol(1e-12)
    pcg.SetMaxIter(200)
    pcg.SetPrintLevel(2)
    pcg.SetPreconditioner(amg)
    pcg.Mult(B, X)
else:
    gmres = mfem.GMRESSolver(MPI.COMM_WORLD)
    gmres.SetAbsTol(0.0)
    gmres.SetRelTol(1e-12)
    gmres.SetMaxIter(200)
    gmres.SetKDim(10)
    gmres.SetPrintLevel(1)
    gmres.SetOperator(A)
    gmres.SetPreconditioner(amg)
    gmres.Mult(B, X)
del amg

# 12. Extract the parallel grid function corresponding to the finite element
#     approximation X. This is the local solution on each processor.
x.Assign(X)

# 13. Save the refined mesh and the solution in parallel. This output can
#     be viewed later using GLVis: "glvis -np <np> -m mesh -g sol".
smyid = '{:0>6d}'.format(myid)
mesh_name = "mesh."+smyid
sol_name = "sol."+smyid

pmesh.Print(mesh_name, 8)
x.Save(sol_name, 8)

# 14. Send the solution by socket to a GLVis server.
if (visualization):
    sol_sock = mfem.socketstream("localhost", 19916)
    sol_sock.send_text("parallel " + str(num_procs) + " " + str(myid))
    sol_sock.precision(8)
    sol_sock.send_solution(pmesh,  x)



from __future__ import print_function
import os
import sys
import numpy as np

if len(sys.argv) > 1 and sys.argv[1] == '-p':   
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint as print
    from mpi4py import MPI
    myid  = MPI.COMM_WORLD.rank
    
else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0
    
def run_test():
    a = mfem.intArray()
    assert a.Size() == 0, "array size is wrong"

    mt = mfem.MemoryType.HOST
    a = mfem.intArray(mt)
    assert a.Size() == 0, "array size is wrong"

    mt = mfem.MemoryType.HOST
    a = mfem.intArray(3, mt)
    assert a.Size() == 3, "array size is wrong"

    
    a = mfem.intArray([1,2,3])
    assert a.Size() == 3, "array size is wrong"
    for i in range(a.Size()):
        assert a[i] == i + 1, "wrong data is set"

    a.Print()
    a.Print("intArray.dat")
    a.Save("intArray_save.dat")

    a2 = mfem.intArray([a.GetData(), a.Size()])
    a3 = mfem.intArray((a.GetData(), a.Size()))    

    a.ToList()
    a[2] = 4
    assert a2.ToList() == a3.ToList(), "pointer constructer fails"
    
    b = mfem.doubleArray([1.1,2.2,3.])
    for i in range(b.Size()):
        print(b[i])
    b.Print()
    b.Print("doubleArray.dat")
    b.Save("doubleArray_save.dat")

    b2 = mfem.doubleArray([b.GetData(), b.Size()])
    b3 = mfem.doubleArray((b.GetData(), b.Size()))
    b[1] = 4    
    print(b2.ToList())
    print(b3.ToList())    

    a = mfem.boolArray([True]*3)
    print(a[1])
    print(a.ToList())    
    assert a.Size() == 3, "array size is wrong"
    
    
if __name__=='__main__':
    run_test()
 



from __future__ import print_function
from numba import cfunc, farray, carray
import os
from os.path import expanduser, join
import sys
import io
from mfem import path as mfem_path
import numpy as np
import numba
import time


if len(sys.argv) > 1 and sys.argv[1] == '-p':
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint
    from mpi4py import MPI
    myid = MPI.COMM_WORLD.rank

else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0
    nicePrint = print


class s_coeff(mfem.PyCoefficient):
    def __init__(self):
        mfem.PyCoefficient.__init__(self)

    def EvalValue(self, p):
        return p[0]


class v_coeff(mfem.VectorPyCoefficient):
    def __init__(self, dim):
        mfem.VectorPyCoefficient.__init__(self, dim)

    def EvalValue(self, p):
        return (p[0], p[1], p[2])


class m_coeff(mfem.MatrixPyCoefficient):
    def __init__(self, dim):
        mfem.MatrixPyCoefficient.__init__(self, dim)

    def EvalValue(self, p):
        return np.array([[p[0], p[1], p[2]],
                         [0.0, p[1], p[2]],
                         [0.0, 0.0, p[2]]])


@cfunc("float64(float64, float64, float64)")
def s_func0(x, y, z):
    return x


def s_func1(x, y, z):
    return x


s_func0 = cfunc("float64(float64, float64, float64)")(s_func1)


@cfunc(mfem.scalar_sig, cache=False)
def s_func(ptx, sdim):
    return s_func0(ptx[0], ptx[1],  ptx[2])


@cfunc(mfem.vector_sig)
def v_func(ptx, out, sdim, vdim):
    out_array = carray(out, (vdim, ))
    for i in range(sdim):
        out_array[i] = ptx[i]


@cfunc(mfem.matrix_sig)
def m_func(ptx, out, sdim, vdim):
    # we use farray to assign the data like above.
    # note out is zero-ed in wrapper. so we don't need to
    # set zero here.
    out_array = farray(out, (vdim, vdim))
    out_array[0, 0] = ptx[0]
    out_array[0, 1] = ptx[1]
    out_array[0, 2] = ptx[2]
    #out_array[1, 0] = 0.0
    out_array[1, 1] = ptx[1]
    out_array[1, 2] = ptx[2]
    #out_array[2, 0] = 0.0
    #out_array[2, 1] = 0.0
    out_array[2, 2] = ptx[2]
    '''
    accessing the array linearly does not speed up.
    out_array = carray(out, (vdim * vdim, ))
    out_array[0] = ptx[0]
    out_array[1] = 0.0
    out_array[2] = 0.0
    out_array[3] = ptx[1]
    out_array[4] = ptx[1]
    out_array[5] = 0.0
    out_array[6] = ptx[2]
    out_array[7] = ptx[2]
    out_array[8] = ptx[2]
    '''

# if dim is know, this provides a simpler way to use matrix coefficient

d1 = mfem.ConstantCoefficient(3)
d2 = mfem.VectorConstantCoefficient([1,2])
d3 = mfem.MatrixConstantCoefficient([[30,40], [50, 60]])
@mfem.jit.matrix(shape=(3,3), sdim=3, dependency=(d1,d2, d3, (d2,d2), d1))
def m_func2(p,d1, d2, d3, d4, d5):
#@mfem.jit.matrix(shape=(3,3), sdim=3)
#def m_func2(p):
    #out_array = farray(out, (3, 3))
    #out_array[0, 0] = ptx[0]
    #out_array[0, 1] = ptx[1]
    #out_array[0, 2] = ptx[2]
    #out_array[1, 1] = ptx[1]
    #out_array[1, 2] = ptx[2]
    #out_array[2, 2] = ptx[2]
    print(d2[0], d2[1], d1, d3[0,0], d3[0,1], d3[1, 0], d3[1,1], d4[0], d4[1], d5)
    return np.array([[p[0], p[1], p[2]],
                     [0.0, p[1], p[2]],
                     [0.0, 0.0, p[2]]])


def check(a, b, msg):
    assert len(a) == len(b), msg
    assert np.sum(np.abs(a - b)) == 0, msg


def run_test():
    #meshfile = expanduser(join(mfem_path, 'data', 'semi_circle.mesh'))
    mesh = mfem.Mesh(3, 3, 3, "TETRAHEDRON")
    mesh.ReorientTetMesh()

    order = 1

    dim = mesh.Dimension()
    sdim = mesh.SpaceDimension()
    fec1 = mfem.H1_FECollection(order, dim)
    fespace1 = mfem.FiniteElementSpace(mesh, fec1, 1)

    fec2 = mfem.ND_FECollection(order, dim)
    fespace2 = mfem.FiniteElementSpace(mesh, fec2, 1)

    print("Element order :", order)

    print('Number of H1 finite element unknowns: ' +
          str(fespace1.GetTrueVSize()))
    print('Number of ND finite element unknowns: ' +
          str(fespace2.GetTrueVSize()))

    print("Checking scalar")

    gf = mfem.GridFunction(fespace1)
    c1 = mfem.NumbaFunction(s_func, sdim).GenerateCoefficient()

    @mfem.jit.scalar
    def c11(ptx):
        return s_func0(ptx[0], ptx[1],  ptx[2])

    @mfem.jit.scalar(td=True, debug=True)
    def c12(ptx, t):
        return s_func0(ptx[0], ptx[1],  ptx[2])

    c2 = s_coeff()

    gf.Assign(0.0)
    start = time.time()
    gf.ProjectCoefficient(c12)
    end = time.time()
    data1 = gf.GetDataArray().copy()
    print("Numba time (scalar)", end - start)

    gf.Assign(0.0)
    start = time.time()
    gf.ProjectCoefficient(c2)
    end = time.time()
    data2 = gf.GetDataArray().copy()
    print("Python time (scalar)", end - start)

    check(data1, data2, "scalar coefficient does not agree with original")

    print("Checking vector")
    gf = mfem.GridFunction(fespace2)
    c3 = mfem.VectorNumbaFunction(v_func, sdim, dim).GenerateCoefficient()
    c4 = v_coeff(dim)

    @mfem.jit.vector(vdim=3, dependency=(c3, c11), td=True,  complex=True)
    def v_func4(ptx, t, c3, c11):
        return np.array([c3[0], c3[1], c3[2]], dtype=np.complex128)
    # @mfem.jit.vector(sdim=3, complex=True)
    # def v_func4(ptx,  ):
    #    return np.array([ptx[0],ptx[1],ptx[2]], dtype=np.complex128)

    gf.Assign(0.0)
    start = time.time()
    for i in range(10):
        gf.ProjectCoefficient(c3)
    end = time.time()
    data1 = gf.GetDataArray().copy()
    print("Numba time (vector)", end - start)

    gf.Assign(0.0)
    start = time.time()
    for i in range(10):
        gf.ProjectCoefficient(c4)
    end = time.time()
    data2 = gf.GetDataArray().copy()
    print("Python time (vector)", end - start)

    gf.Assign(0.0)
    start = time.time()
    for i in range(10):
        gf.ProjectCoefficient(v_func4.real)
    end = time.time()
    data3 = gf.GetDataArray().copy()
    print("Numba2 time (vector)", end - start)

    check(data1, data2, "vector coefficient does not agree with original")
    check(data1, data3, "vector coefficient does not agree with original")

    print("speed comparision with C++")

    @mfem.jit.vector(vdim=3, interface="c++", debug=True)
    def v_func4_old(ptx, out):
        out[0] = 1
        out[1] = 2.
        out[2] = 3

    @mfem.jit.vector(vdim=3)
    def v_func4_new(ptx):
        return np.array([1, 2, 3.])

    gf.Assign(0.0)
    start = time.time()
    for i in range(10):
        gf.ProjectCoefficient(v_func4_old)
    end = time.time()
    data3 = gf.GetDataArray().copy()
    print("Numba time (vector) - old interface", end - start)

    gf.Assign(0.0)
    start = time.time()
    for i in range(10):
        gf.ProjectCoefficient(v_func4_new)
    end = time.time()
    data3 = gf.GetDataArray().copy()
    print("Numba time (vector) - new interface", end - start)

    val = mfem.Vector([1, 2, 3])
    cc = mfem.VectorConstantCoefficient(val)
    gf.Assign(0.0)
    start = time.time()
    for i in range(10):
        gf.ProjectCoefficient(cc)
    end = time.time()
    data3 = gf.GetDataArray().copy()
    print("C++ constant coefficient", end - start)

    print("Checking matrix")
    a1 = mfem.BilinearForm(fespace2)
    a2 = mfem.BilinearForm(fespace2)
    a3 = mfem.BilinearForm(fespace2)
    a4 = mfem.BilinearForm(fespace2)
    a5 = mfem.BilinearForm(fespace2)

    c4 = mfem.MatrixNumbaFunction(m_func, sdim, dim).GenerateCoefficient()
    c5 = m_coeff(dim)

    a1.AddDomainIntegrator(mfem.VectorFEMassIntegrator(c4))
    a2.AddDomainIntegrator(mfem.VectorFEMassIntegrator(c5))
    a3.AddDomainIntegrator(mfem.VectorFEMassIntegrator(m_func2))

    @mfem.jit.matrix(shape=(3,3), dependency=(c4, c4), complex=True, td=True)
    def m_func3(ptx, t, c4, c5):
        ret = np.array([[c5[0, 0], c5[0, 1], c5[0, 2]],
                        [0.0,      c5[1, 1], c5[1, 2]],
                        [0.0,      0.0,      c5[2, 2]]])
        return (t/3.0)*ret*1j

    @mfem.jit.matrix(shape=(3,3), dependency=(m_func3, c4))
    def m_func4_complex(ptx, m_func3, c5):
        return m_func3.imag

    @mfem.jit.matrix(shape=(3,3), dependency=((m_func3.real, m_func3.imag), c4), td=True)
    def m_func4_split(ptx, t, m_func3, c5):
        return m_func3.imag*(t/3.0)

    '''
    @mfem.jit.matrix(sdim=3, complex=False, debug=True)
    def m_func5(p):
        x = p[0]
        y = p[1]
        return np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
    '''
    # _out_ =
    # return np.array(_out_)

    a4.AddDomainIntegrator(mfem.VectorFEMassIntegrator(m_func4_complex))
    a5.AddDomainIntegrator(mfem.VectorFEMassIntegrator(m_func4_split))

    start = time.time()
    a1.Assemble()
    end = time.time()
    a1.Finalize()
    M1 = a1.SpMat()

    print("Numba time (matrix)", end - start)

    start = time.time()
    a2.Assemble()
    end = time.time()
    a2.Finalize()
    M2 = a2.SpMat()
    print("Python time (matrix)", end - start)

    start = time.time()
    a3.Assemble()
    end = time.time()
    a3.Finalize()
    M3 = a3.SpMat()
    nicePrint("Numba (simpler interface) (matrix)", end - start)

    start = time.time()
    m_func4_complex.SetTime(3.0)
    a4.Assemble()
    end = time.time()
    a4.Finalize()
    M4 = a4.SpMat()
    nicePrint("Numba (complex dependency as complex) (matrix)", end - start)

    start = time.time()
    m_func4_split.SetTime(3.0)
    a5.Assemble()
    end = time.time()
    a5.Finalize()
    M5 = a5.SpMat()
    nicePrint("Numba (complex dependency as decomposed) (matrix)", end - start)

    #from mfem.commmon.sparse_utils import sparsemat_to_scipycsr
    #csr1 = sparsemat_to_scipycsr(M1, float)
    #csr2 = sparsemat_to_scipycsr(M2, float)

    def compare_mat(M1o, M2o):
        check(M1o.GetDataArray(),
              M2o.GetDataArray(),
              "matrix coefficient does not agree with original")
        check(M1o.GetIArray(),
              M2o.GetIArray(),
              "matrix coefficient does not agree with original")
        check(M1o.GetJArray(),
              M2o.GetJArray(),
              "matrix coefficient does not agree with original")

    compare_mat(M1, M2)
    compare_mat(M1, M3)
    compare_mat(M1, M4)
    compare_mat(M1, M5)

    # print(m_func3.SpaceDimension())
    nicePrint("PASSED")


if __name__ == '__main__':
    run_test()



'''
   MFEM example 28
      See c++ version in the MFEM library for more detail 
'''
import os
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp, sqrt, pi, abs, array, floor, log


def build_trapezoid_mesh(offset):
    assert offset < 0.9, "offset is too large!"

    dimension = 2
    nvt = 4  # vertices
    nbe = 4  # num boundary elements
    mesh = mfem.Mesh(dimension, nvt, 1, nbe)

    mesh.AddVertex(0.0, 0.0)
    mesh.AddVertex(1.0, 0.0)
    mesh.AddVertex(offset, 1.0)
    mesh.AddVertex(1.0, 1.0)

    #  element
    mesh.AddQuad([0, 1, 3, 2], 1)

    #  boundary
    mesh.AddBdrSegment([0, 1], 1)
    mesh.AddBdrSegment([1, 3], 2)
    mesh.AddBdrSegment([2, 3], 3)
    mesh.AddBdrSegment([0, 2], 4)

    mesh.FinalizeQuadMesh(1, 0, True)

    return mesh


def run(order=1,
        offset=0.3,
        visit=False,
        visualization=True):

    # 2. Build a trapezoidal mesh with a single quadrilateral element, where
    #    'offset' determines how far off it is from a rectangle.
    mesh = build_trapezoid_mesh(offset)
    dim = mesh.Dimension()

    # 3. Refine the mesh to increase the resolution. In this example we do
    #    'ref_levels' of uniform refinement. We choose 'ref_levels' to be the
    #    largest number that gives a final mesh with no more than 1,000
    #    elements.
    ref_levels = int(floor(log(1000./mesh.GetNE())/log(2.)/dim))
    for l in range(ref_levels):
        mesh.UniformRefinement()

    # 4. Define a finite element space on the mesh. Here we use vector finite
    #    elements, i.e. dim copies of a scalar finite element space. The vector
    #    dimension is specified by the last argument of the FiniteElementSpace
    #    constructor.

    fec = mfem.H1_FECollection(order, dim)
    fespace = mfem.FiniteElementSpace(mesh, fec, dim)
    print("Number of finite element unknowns: " + str(fespace.GetTrueVSize()))
    print("Assembling matrix and r.h.s... ")

    # 5. Determine the list of true (i.e. parallel conforming) essential
    #    boundary dofs. In this example, there are no essential boundary
    #    conditions in the usual sense, but we leave the machinery here for
    #    users to modify if they wish.
    ess_tdof_list = mfem.intArray()
    ess_bdr = mfem.intArray(mesh.bdr_attributes.Max())
    ess_bdr.Assign(0)
    fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    # 6. Set up the linear form b(.) which corresponds to the right-hand side of
    #    the FEM linear system. In this case, b_i equals the boundary integral
    #    of f*phi_i where f represents a "push" force on the right side of the
    #    trapezoid.
    f = mfem.VectorArrayCoefficient(dim)
    for i in range(dim-1):
        f.Set(i, mfem.ConstantCoefficient(0.0))
    push_force = mfem.Vector(mesh.bdr_attributes.Max())
    push_force.Assign(0.0)
    push_force[1] = -5.0e-2
    f.Set(0, mfem.PWConstCoefficient(push_force))

    b = mfem.LinearForm(fespace)
    b.AddBoundaryIntegrator(mfem.VectorBoundaryLFIntegrator(f))
    b.Assemble()

    # 7. Define the solution vector x as a finite element grid function
    #    corresponding to fespace.
    x = mfem.GridFunction(fespace)
    x.Assign(0.0)

    # 8. Set up the bilinear form a(.,.) on the finite element space
    #    corresponding to the linear elasticity integrator with piece-wise
    #    constants coefficient lambda and mu. We use constant coefficients,
    #    but see ex2 for how to set up piecewise constant coefficients based
    #    on attribute.
    llambda = mfem.Vector(mesh.attributes.Max())
    llambda.Assign(1.0)
    lambda_func = mfem.PWConstCoefficient(llambda)
    mu = mfem.Vector(mesh.attributes.Max())
    mu.Assign(1.0)
    mu_func = mfem.PWConstCoefficient(mu)

    a = mfem.BilinearForm(fespace)
    a.AddDomainIntegrator(mfem.ElasticityIntegrator(lambda_func, mu_func))

    # 9. Assemble the bilinear form and the corresponding linear system,
    #    applying any necessary transformations such as: eliminating boundary
    #    conditions, applying conforming constraints for non-conforming AMR,
    #    static condensation, etc.
    a.Assemble()

    A = mfem.SparseMatrix()
    B = mfem.Vector()
    X = mfem.Vector()
    a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)
    print("done.")
    print("Size of linear system: " + str(A.Height()))

    # 10. Set up constraint matrix to constrain normal displacement (but
    #     allow tangential displacement) on specified boundaries.
    constraint_atts = mfem.intArray([1, 4])
    lagrange_rowstarts = mfem.intArray()
    local_constraints = mfem.BuildNormalConstraints(fespace,
                                                    constraint_atts,
                                                    lagrange_rowstarts)

    # 11. Define and apply an iterative solver for the constrained system
    #     in saddle-point form with a Gauss-Seidel smoother for the
    #     displacement block.
    M = mfem.GSSmoother(A)
    solver = mfem.SchurConstrainedSolver(A, local_constraints, M)
    solver.SetRelTol(1e-5)
    solver.SetMaxIter(2000)
    solver.SetPrintLevel(1)
    solver.Mult(B, X)

    # 12. Recover the solution as a finite element grid function. Move the
    #     mesh to reflect the displacement of the elastic body being
    #     simulated, for purposes of output.
    a.RecoverFEMSolution(X, b, x)
    mesh.SetNodalFESpace(fespace)
    nodes = mesh.GetNodes()
    nodes += x

    #  13. Save the refined mesh and the solution in VisIt format.
    if visit:
        visit_dc = mfem.VisItDataCollection("ex28", mesh)
        visit_dc.SetLevelsOfDetail(4)
        visit_dc.RegisterField("displacement", x)
        visit_dc.Save()

    # 14. Save the displaced mesh and the inverted solution (which gives the
    #     backward displacements to the original grid). This output can be
    #     viewed later using GLVis: "glvis -m displaced.mesh -g sol.gf".
    x *= -1  # sign convention for GLVis displacements
    mesh.Print("displaced.mesh", 8)
    x.Save("sol.gf", 8)

    # 15. Send the above data by socket to a GLVis server.  Use the "n" and "b"
    #     keys in GLVis to visualize the displacements.
    if visualization:
        sol_sock = mfem.socketstream('localhost', 19916)
        sol_sock.precision(8)
        sol_sock << "solution\n" << mesh << x
        sol_sock.flush


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(
        description='Ex28 (Constraints and sliding boundary conditions')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument('-offset', '--offset',
                        action='store', default=0.3, type=float,
                        help="How much to offset the trapezoid.")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')
    parser.add_argument('-visit', '--visit-datafiles',
                        action='store_true',
                        help="Save data files for VisIt (visit.llnl.gov) visualization.")

    args = parser.parse_args()
    parser.print_options(args)

    run(order=args.order,
        offset=args.offset,
        visit=args.visit_datafiles,
        visualization=args.visualization,)



'''
   MFEM example 22
      See c++ version in the MFEM library for more detail 
'''
import os
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp, sqrt, pi


def run(mesh_file="",
        order=1,
        ref_levels=0,
        visualization=1,
        prob=0,
        harm_conv=True,
        a_coef=0.0,
        epsilon=1.0,
        sigma=20.,
        mu=1.0,
        device='cpu',
        pa=False,
        freq=-1.0,
        numba=False):

    if a_coef != 0.0:
        mu = 1.0 / a_coef
    omega = 10.
    if freq > 0.0:
        omega = 2.0 * pi * freq
    exact_sol = check_for_inline_mesh(mesh_file)
    if exact_sol:
        print("Identified a mesh with known exact solution")

    conv = mfem.ComplexOperator.HERMITIAN if harm_conv else mfem.ComplexOperator.BLOCK_SYMMETRIC

    # 2. Enable hardware devices such as GPUs, and programming models such as
    #    CUDA, OCCA, RAJA and OpenMP based on command line options.
    device = mfem.Device(device)
    device.Print()

    # 3. Read the mesh from the given mesh file. We can handle triangular,
    #    quadrilateral, tetrahedral, hexahedral, surface and volume meshes
    #    with the same code.
    mesh = mfem.Mesh(mesh_file, 1, 1)
    dim = mesh.Dimension()

    # 4. Refine the mesh to increase resolution. In this example we do
    #    'ref_levels' of uniform refinement where the user specifies
    #    the number of levels with the '-r' option.
    for i in range(ref_levels):
        mesh.UniformRefinement()

    # 5. Define a finite element space on the mesh. Here we use continuous
    #    Lagrange, Nedelec, or Raviart-Thomas finite elements of the specified
    #    order.
    if (dim == 1 and prob != 0):
        print("Switching to problem type 0, H1 basis functions, for 1 dimensional mesh.")
        prob = 0

    if prob == 0:
        fec = mfem.H1_FECollection(order, dim)
    elif prob == 1:
        fec = mfem.ND_FECollection(order, dim)
    elif prob == 2:
        fec = mfem.RT_FECollection(order-1, dim)
    else:
        assert False, "unknown problem"

    fespace = mfem.FiniteElementSpace(mesh, fec)
    print("Number of finite element unknowns: " + str(fespace.GetTrueVSize()))

    # 6. Determine the list of true (i.e. conforming) essential boundary dofs.
    #    In this example, the boundary conditions are defined based on the type
    #    of mesh and the problem type.
    ess_tdof_list = mfem.intArray()
    ess_bdr = mfem.intArray()

    if mesh.bdr_attributes.Size():
        ess_bdr.SetSize(mesh.bdr_attributes.Max())
        ess_bdr.Assign(1)
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    # 7. Set up the linear form b(.) which corresponds to the right-hand side of
    #    the FEM linear system.
    b = mfem.ComplexLinearForm(fespace, conv)
    b.Assign(0.0)

    # 8. Define the solution vector u as a complex finite element grid function
    #    corresponding to fespace. Initialize u with initial guess of 1+0i or
    #    the exact solution if it is known.

    # 8-a. Prepare coefficient in pure Python and numba
    sdim = mesh.SpaceDimension()

    if numba:
        @mfem.jit.scalar()
        def u0_real_exact(x):
            alpha = (epsilon * omega - 1j * sigma)
            kappa = sqrt(mu * omega * alpha)
            return exp(-1j * kappa * x[dim - 1]).real

        @mfem.jit.scalar()
        def u0_imag_exact(x):
            alpha = (epsilon * omega - 1j * sigma)
            kappa = sqrt(mu * omega * alpha)
            return exp(-1j * kappa * x[dim - 1]).imag

        # mfem.jit.vector can either turn numpy array (defualt) or pass
        # the result using the second argment (interface="c++")
        @mfem.jit.vector(vdim=sdim)
        def u1_real_exact(x):
            alpha = (epsilon * omega - 1j * sigma)
            kappa = sqrt(mu * omega * alpha)
            return np.array([exp(-1j * kappa * x[dim - 1]).real, 0.0, 0.0])

        @mfem.jit.vector(vdim=sdim)
        def u1_imag_exact(x):
            alpha = (epsilon * omega - 1j * sigma)
            kappa = sqrt(mu * omega * alpha)
            return np.array([exp(-1j * kappa * x[dim - 1]).imag, 0.0, 0.0])

        @mfem.jit.vector(vdim=sdim, interface="c++")
        def u2_real_exact(x, out):
            alpha = (epsilon * omega - 1j * sigma)
            kappa = sqrt(mu * omega * alpha)
            for i in range(dim):
                out[i] = 0
            out[dim-1] = exp(-1j * kappa * x[dim - 1]).real

        @mfem.jit.vector(vdim=sdim, interface="c++")
        def u2_imag_exact(x, out):
            alpha = (epsilon * omega - 1j * sigma)
            kappa = sqrt(mu * omega * alpha)
            for i in range(dim):
                out[i] = 0
            out[dim-1] = exp(-1j * kappa * x[dim - 1]).imag

    else:
        def u0_exact(x):
            alpha = (epsilon * omega - 1j * sigma)
            kappa = sqrt(mu * omega * alpha)
            return exp(-1j * kappa * x[dim - 1])

        class cu0_real_exact(mfem.PyCoefficient):
            def EvalValue(self, x):
                return u0_exact(x).real

        class cu0_imag_exact(mfem.PyCoefficient):
            def EvalValue(self, x):
                return u0_exact(x).imag
        u0_real_exact = cu0_real_exact()
        u0_imag_exact = cu0_imag_exact()

        class cu1_real_exact(mfem.VectorPyCoefficient):
            def __init__(self):
                mfem.VectorPyCoefficient.__init__(self, sdim)

            def EvalValue(self, x):
                ret = x*0
                ret[0] = u0_exact(x).real
                return ret

        class cu1_imag_exact(mfem.VectorPyCoefficient):
            def __init__(self):
                mfem.VectorPyCoefficient.__init__(self, sdim)

            def EvalValue(self, x):
                ret = x*0
                ret[0] = u0_exact(x).imag
                return ret

        u1_real_exact = cu1_real_exact()
        u1_imag_exact = cu1_imag_exact()

        class cu2_real_exact(mfem.VectorPyCoefficient):
            def __init__(self):
                mfem.VectorPyCoefficient.__init__(self, sdim)

            def EvalValue(self, x):
                ret = x*0
                ret[dim-1] = u0_exact(x).real
                return ret

        class cu2_imag_exact(mfem.VectorPyCoefficient):
            def __init__(self):
                mfem.VectorPyCoefficient.__init__(self, sdim)

            def EvalValue(self, x):
                ret = x*0
                ret[dim-1] = u0_exact(x).imag
                return ret

        u2_real_exact = cu2_real_exact()
        u2_imag_exact = cu2_imag_exact()

    u = mfem.ComplexGridFunction(fespace)
    if exact_sol:
        u_exact = mfem.ComplexGridFunction(fespace)

    u0_r = u0_real_exact
    u0_i = u0_imag_exact
    u1_r = u1_real_exact
    u1_i = u1_imag_exact
    u2_r = u2_real_exact
    u2_i = u2_imag_exact

    zeroCoef = mfem.ConstantCoefficient(0.0)
    oneCoef = mfem.ConstantCoefficient(1.0)

    zeroVec = mfem.Vector(dim)
    zeroVec.Assign(0.0)
    oneVec = mfem.Vector(dim)
    oneVec.Assign(0.0)
    oneVec[dim-1 if prob == 2 else 0] = 1.0
    zeroVecCoeff = mfem.VectorConstantCoefficient(zeroVec)
    oneVecCoeff = mfem.VectorConstantCoefficient(oneVec)

    if prob == 0:
        if exact_sol:
            u.ProjectBdrCoefficient(u0_r, u0_i, ess_bdr)
            u_exact.ProjectCoefficient(u0_r, u0_i)
        else:
            u.ProjectBdrCoefficient(oneCoef, zeroCoef, ess_bdr)
    elif prob == 1:
        if exact_sol:
            u.ProjectBdrCoefficientTangent(u1_r, u1_i, ess_bdr)
            u_exact.ProjectCoefficient(u1_r, u1_i)
        else:
            u.ProjectBdrCoefficientTangent(oneVecCoef, zeroVecCoef, ess_bdr)
    elif prob == 2:
        if exact_sol:
            u.ProjectBdrCoefficientNormal(u2_r, u2_i, ess_bdr)
            u_exact.ProjectCoefficient(u2_r, u2_i)
        else:
            u.ProjectBdrCoefficientNormal(oneVecCoef, zeroVecCoef, ess_bdr)

    if (visualization and exact_sol):
        sol_sock_r = mfem.socketstream("localhost", 19916)
        sol_sock_r.precision(8)
        sol_sock_r << "solution\n" << mesh << u_exact.real(
        ) << "window_title 'Exact: Real Part'"
        sol_sock_r.flush()
        sol_sock_i = mfem.socketstream("localhost", 19916)
        sol_sock_i.precision(8)
        sol_sock_i << "solution\n" << mesh << u_exact.imag(
        ) << "window_title 'Exact: Imag Part'"
        sol_sock_i.flush()

    # 9. Set up the sesquilinear form a(.,.) on the finite element space
    #    corresponding to the damped harmonic oscillator operator of the
    #    appropriate type:
    #
    #    0) A scalar H1 field
    #       -Div(a Grad) - omega^2 b + i omega c
    #
    #    1) A vector H(Curl) field
    #       Curl(a Curl) - omega^2 b + i omega c
    #
    #    2) A vector H(Div) field
    #       -Grad(a Div) - omega^2 b + i omega c

    stiffnessCoef = mfem.ConstantCoefficient(1.0/mu)
    massCoef = mfem.ConstantCoefficient(-omega * omega * epsilon)
    lossCoef = mfem.ConstantCoefficient(omega * sigma)
    negMassCoef = mfem.ConstantCoefficient(omega * omega * epsilon)

    a = mfem.SesquilinearForm(fespace, conv)
    if pa:
        a.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)
    if prob == 0:
        a.AddDomainIntegrator(mfem.DiffusionIntegrator(stiffnessCoef),
                              None)
        a.AddDomainIntegrator(mfem.MassIntegrator(massCoef),
                              mfem.MassIntegrator(lossCoef))
    elif prob == 1:
        a.AddDomainIntegrator(mfem.CurlCurlIntegrator(stiffnessCoef),
                              None)
        a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(massCoef),
                              mfem.VectorFEMassIntegrator(lossCoef))
    elif prob == 2:
        a.AddDomainIntegrator(mfem.DivDivIntegrator(stiffnessCoef),
                              None)
        a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(massCoef),
                              mfem.VectorFEMassIntegrator(lossCoef))
    else:
        assert False, "Unknown probm"

    # 9a. Set up the bilinear form for the preconditioner corresponding to the
    #     appropriate operator
    #
    #      0) A scalar H1 field
    #         -Div(a Grad) - omega^2 b + omega c
    #
    #      1) A vector H(Curl) field
    #         Curl(a Curl) + omega^2 b + omega c
    #
    #      2) A vector H(Div) field
    #         -Grad(a Div) - omega^2 b + omega c

    pcOp = mfem.BilinearForm(fespace)
    if pa:
        pcOp.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)

    if prob == 0:
        pcOp.AddDomainIntegrator(mfem.DiffusionIntegrator(stiffnessCoef))
        pcOp.AddDomainIntegrator(mfem.MassIntegrator(massCoef))
        pcOp.AddDomainIntegrator(mfem.MassIntegrator(lossCoef))
    elif prob == 1:
        pcOp.AddDomainIntegrator(mfem.CurlCurlIntegrator(stiffnessCoef))
        pcOp.AddDomainIntegrator(mfem.VectorFEMassIntegrator(negMassCoef))
        pcOp.AddDomainIntegrator(mfem.VectorFEMassIntegrator(lossCoef))
    elif prob == 2:
        pcOp.AddDomainIntegrator(mfem.DivDivIntegrator(stiffnessCoef))
        pcOp.AddDomainIntegrator(mfem.VectorFEMassIntegrator(massCoef))
        pcOp.AddDomainIntegrator(mfem.VectorFEMassIntegrator(lossCoef))

    # 10. Assemble the form and the corresponding linear system, applying any
    #     necessary transformations such as: assembly, eliminating boundary
    #     conditions, conforming constraints for non-conforming AMR, etc.
    a.Assemble()
    pcOp.Assemble()

    A = mfem.OperatorHandle()
    B = mfem.Vector()
    U = mfem.Vector()
    a.FormLinearSystem(ess_tdof_list, u, b, A, U, B)

    print("Size of linear system: " + str(A.Width()))

    # 11. Define and apply a GMRES solver for AU=B with a block diagonal
    #     preconditioner based on the appropriate sparse smoother.
    blockOffsets = mfem.intArray()
    blockOffsets.SetSize(3)
    blockOffsets[0] = 0
    blockOffsets[1] = A.Height() // 2
    blockOffsets[2] = A.Height() // 2
    blockOffsets.PartialSum()

    BDP = mfem.BlockDiagonalPreconditioner(blockOffsets)

    if pa:
        pc_r = mfem.OperatorJacobiSmoother(pcOp, ess_tdof_list)
        pc_i = None
    else:
        PCOp = mfem.OperatorHandle()
        pcOp.SetDiagonalPolicy(mfem.Operator.DIAG_ONE)
        pcOp.FormSystemMatrix(ess_tdof_list, PCOp)
        if prob == 0:
            pc_r = mfem.DSmoother(mfem.OperatorHandle2SparseMatrix(PCOp))
        elif prob == 1:
            pc_r = mfem.GSSmoother(mfem.OperatorHandle2SparseMatrix(PCOp))
        elif prob == 2:
            pc_r = mfem.DSmoother(mfem.OperatorHandle2SparseMatrix(PCOp))

    s = 1.0 if prob != 1 else -1.0

    pc_i = mfem.ScaledOperator(pc_r,
                               s if conv == mfem.ComplexOperator.HERMITIAN else -s)
    BDP.SetDiagonalBlock(0, pc_r)
    BDP.SetDiagonalBlock(1, pc_i)
    # BDP.owns_blocks = 1  #this is not necessary since Python owns smoothers

    gmres = mfem.GMRESSolver()
    gmres.SetPreconditioner(BDP)
    gmres.SetOperator(A.Ptr())
    gmres.SetRelTol(1e-12)
    gmres.SetMaxIter(1000)
    gmres.SetPrintLevel(1)
    gmres.Mult(B, U)

    # 12. Recover the solution as a finite element grid function and compute the
    #     errors if the exact solution is known.
    a.RecoverFEMSolution(U, b, u)

    if exact_sol:
        err_r = -1.0
        err_i = -1.0

        if prob == 0:
            err_r = u.real().ComputeL2Error(u0_r)
            err_i = u.imag().ComputeL2Error(u0_i)
        elif prob == 1:
            err_r = u.real().ComputeL2Error(u1_r)
            err_i = u.imag().ComputeL2Error(u1_i)
        elif prob == 2:
            err_r = u.real().ComputeL2Error(u2_r)
            err_i = u.imag().ComputeL2Error(u2_i)

    print()
    print("|| Re (u_h - u) ||_{L^2} = " + "{:g}".format(err_r))
    print("|| Im (u_h - u) ||_{L^2} = " + "{:g}".format(err_i))

    # 13. Save the refined mesh and the solution. This output can be viewed
    #     later using GLVis: "glvis -m mesh -g sol".
    mesh.Print("refined.mesh", 8)
    u.real().Save("sol_r.gf", 8)
    u.imag().Save("sol_i.gf", 8)

    # 14. Send the solution by socket to a GLVis server.
    if visualization:
        sol_sock_r = mfem.socketstream("localhost", 19916)
        sol_sock_r.precision(8)
        sol_sock_r << "solution\n" << mesh << u.real(
        ) << "window_title 'Solution: Real Part'"
        sol_sock_i = mfem.socketstream("localhost", 19916)
        sol_sock_i.precision(8)
        sol_sock_i << "solution\n" << mesh << u.imag(
        ) << "window_title 'Solution: Imag Part'"

    if visualization and exact_sol:
        u_exact -= u
        sol_sock_r = mfem.socketstream("localhost", 19916)
        sol_sock_r.precision(8)
        sol_sock_r << "solution\n" << mesh << u_exact.real(
        ) << "window_title 'Error: Real Part'"
        sol_sock_i = mfem.socketstream("localhost", 19916)
        sol_sock_i.precision(8)
        sol_sock_i << "solution\n" << mesh << u_exact.imag(
        ) << "window_title 'Error: Imag Part'"

    if visualization:
        u_t = mfem.GridFunction(fespace)
        u_t.Assign(u.real())
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock.precision(8)
        sol_sock << "solution\n" << mesh << u_t
        sol_sock << "window_title 'Harmonic Solution (t = 0.0 T)'"
        sol_sock << "pause\n"
        sol_sock.flush()

        print("GLVis visualization paused. Press space (in the GLVis window) to resume it.")
        num_frames = 32
        i = 0

        # Let's plot one wave cycle...
        for i in range(32):
            t = (i % num_frames) / num_frames
            oss = "Harmonic Solution (t = " + str(t) + " T)"
            dd = (cos(2.0 * pi * t)*u.real().GetDataArray() +
                  sin(2.0 * pi * t)*u.imag().GetDataArray())
            u_t.Assign(mfem.Vector())
            sol_sock << "solution\n" << mesh << u_t
            sol_sock << "window_title '" << oss << "'"
            sol_sock.flush()


def check_for_inline_mesh(mesh_file):
    return os.path.basename(mesh_file)[:7] == "inline-"


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(
        description='Ex22 (Complex linear systems)')
    parser.add_argument('-m', '--mesh',
                        default='inline-quad.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument("-p",
                        "--problem-type",
                        action='store', type=int, default=0,
                        help="\n".join(["Choose between 0: H_1, 1: H(Curl), or 2: H(Div) "
                                        "damped harmonic oscillator."]))
    parser.add_argument("-r", "--refine",
                        action='store', type=int, default=0,
                        help="Number of times to refine the mesh uniformly.")
    parser.add_argument("-a", "--stiffness-coef",
                        action='store', type=float, default=0.0,
                        help="Stiffness coefficient (spring constant or 1/mu).")
    parser.add_argument("-mu", "--permeability",
                        action='store', type=float, default=1.0,
                        help="Permeability of free space (or 1/(spring constant)).")
    parser.add_argument("-eps", "--permittivity",
                        action='store', type=float, default=1.0,
                        help="Permittivity of free space (or mass constant).")
    parser.add_argument("-sigma", "--conductivity",
                        action='store', type=float, default=20.0,
                        help="Conductivity (or damping constant).")
    parser.add_argument("-f", "--frequency",
                        action='store',
                        type=float,
                        default=-1.0,
                        help="Set the frequency for the exact")
    parser.add_argument("-no-herm", "--no-hermitian",
                        action='store_false',
                        default=True,
                        help="Do not use convention for Hermitian operators.")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        default=True,
                        help='Enable GLVis visualization')
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")

    try:
        from numba import jit
        HAS_NUMBA = True
    except ImportError:
        HAS_NUMBA = False
    parser.add_argument("-n", "--numba",
                        default=int(HAS_NUMBA),
                        type=int,
                        action='store',
                        help="Use Number compiled coefficient")

    args = parser.parse_args()
    args.numba = bool(args.numba)
    parser.print_options(args)

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))

    run(mesh_file=meshfile,
        order=args.order,
        ref_levels=args.refine,
        prob=args.problem_type,
        visualization=args.visualization,
        # visualization=False,
        harm_conv=args.no_hermitian,
        a_coef=args.stiffness_coef,
        epsilon=args.permittivity,
        sigma=args.conductivity,
        mu=args.permeability,
        device=args.device,
        pa=args.partial_assembly,
        freq=args.frequency,
        numba=args.numba)



from mpi4py import MPI
import traceback
def get_mpi_datatype(data):
    '''
    get MPI datatype from numpy dtype
    '''
    #print '!!!chekcing MPI type ',  data.dtype.name
    s = data.dtype.itemsize  # get data size of element
    if data.dtype.name.startswith('int'):
        if s == 8:
           return MPI.LONG
        elif s == 4:
           return MPI.INT            
        elif s == 2:         
           return MPI.SHORT
        else:
           pass
    elif data.dtype.name.startswith('float'):       
        if s == 16:
           return MPI.LONG_DOUBLE
        elif s == 8:
           return MPI.DOUBLE
        elif s == 4:
           return MPI.FLOAT
        else:
           pass
    elif data.dtype.name.startswith('complex'):              
        if s == 16:
           return MPI.DOUBLE_COMPLEX
        elif s == 8:
           return MPI.COMPLEX
        else:
           pass
    elif data.dtype.name.startswith('bool'):
        return MPI.BOOL        
    else:
        pass
    assert False, "MPI data type is unknown for " + data.dtype.name
